using System;
using System.Collections.Generic;
using System.Text;
using MinorShift.Emuera.Sub;
using System.Text.RegularExpressions;
using MinorShift.Emuera.GameData.Variable;
using MinorShift.Emuera.GameData.Expression;
using MinorShift.Emuera.GameView;
using MinorShift.Emuera.GameData;

namespace MinorShift.Emuera.GameProc
{
    internal delegate void PrintWarning(string str, LogicalLine line, int level, bool isError);
    internal static class LogicalLineParser
    {
        readonly static Regex regexCom = new Regex("COM[0-9]+");
        readonly static Regex regexComAble = new Regex("COM_ABLE[0-9]+");
        readonly static Regex regexAblup = new Regex("ABLUP[0-9]+");
        private static bool isSystemLabel(string labelName)
        {
            switch (labelName)
            {
                case "EVENTFIRST":
                case "EVENTTRAIN":
                case "EVENTSHOP":
                case "EVENTBUY":
                case "EVENTCOM":
                case "EVENTTURNEND":
                case "EVENTCOMEND":
                case "EVENTEND":
                case "SHOW_STATUS":
                case "SHOW_USERCOM":
                case "USERCOM":
                case "SOURCE_CHECK":
                case "SHOW_JUEL":
                case "SHOW_ABLUP_SELECT":
                case "USERABLUP":
                case "SHOW_SHOP":
                case "SAVEINFO":
				case "USERSHOP":

				case "EVENTLOAD":
				case "TITLE_LOADGAME":
				case "SYSTEM_AUTOSAVE":
				case "SYSTEM_TITLE":
                    return true;
            }

            if (labelName.StartsWith("COM"))
            {
                if (regexCom.IsMatch(labelName))
                    return true;
                if (regexComAble.IsMatch(labelName))
                    return true;
            }
            if (labelName.StartsWith("ABLUP"))
                if (regexAblup.IsMatch(labelName))
                    return true;
            return false;
        }

		readonly static char[] badSymbolAsLabelname = new char[] {
		    '+', '-', '*', '/', '%', '=', '!', '<', '>', '|', '&', '^', '~',
		    ' ', '@', '\t' ,
		    '\"','(', ')', '{', '[', '}', ']', ',', ':',
		    '\\', '@', '$', '#', '?',
		    //'_'OK
		};

        public static LogicalLine ParseLine(string str, EmueraConsole console)
        {
            ScriptPosition position = new ScriptPosition(str);
            StringStream stream = new StringStream(str);
            return ParseLine(stream, position, console);
        }

        public static LogicalLine ParseLine(StringStream stream, ScriptPosition position, EmueraConsole console)
        {
            int lineNo = position.LineNo;
            string rowLine = stream.Substring();
            string s1 = null;
            string s2 = null;
            string s3 = null;
            string errMes = "";
            TokenReader.SkipWhiteSpace(stream);//擪̃zCgXy[Xǂݔ΂
            if (stream.EOS)
                return null;
            //Rgsǂ͂ɗOɔ肵Ă
            //x
            if ((stream.Current == '$') || (stream.Current == '@'))
            {
                bool isFunction = (stream.Current == '@');
                stream.ShiftNext();//@$
                string labelName = TokenReader.ReadStringEndWith(stream, null);
                List<VariableToken> args = new List<VariableToken>();
                VariableToken vars;
				labelName = labelName.Trim();
				if (Config.Instance.IgnoreCase)
					labelName = labelName.ToUpper();
                if (labelName.Length == 0)
                {
                    errMes = "x܂";
                    goto err;
				}
				////Eramakerł̓ubNXgɂL͑Sʂ̂Ō݊mۂ̂߂ȂƂ
				//ϐɎgȂgLOCAL@֐ō̂ŕB
				//eramakerł͐ɓEmuerałʏ͖Ȃ삷̂ŌxLv0ɒቺB
				if (labelName.IndexOfAny(badSymbolAsLabelname) >= 0)
					console.PrintWarning("x" + labelName + "\"_\"ȊŐL܂܂Ă܂B", position, 0);

				if (!stream.EOS)
				{
					if (!isFunction)
					{
						console.PrintWarning("$Ŏn܂郉xɈݒ肳Ă܂", position, 1);
					}
					else
					{
						while (!stream.EOS)
						{
							if (stream.Current != ',')
							{
								errMes = "̏ԈĂ܂";
								goto err;
							}
							stream.ShiftNext();
							TokenReader.SkipWhiteSpace(stream);
							vars = VariableParser.ReduceVariable(stream);
							args.Add(vars);
							TokenReader.SkipWhiteSpace(stream);
						}
					}
				}
                if (isFunction)
                {
                    FunctionLabelLine labelLine = new FunctionLabelLine(position, labelName, args);
                    if (isSystemLabel(labelName))
                        labelLine.Depth = 0;
                    return labelLine;
                }
                else
                {
                    return new GotoLabelLine(position, labelName);
                }
            }
            s1 = TokenReader.ReadIdentiferWithIndex(stream);
            BuiltInFunctionCode func = BuiltInFunctionManager.ToBuiltInFunction(s1);
            //ߕ
            if (func != BuiltInFunctionCode.__NULL__)//֐
            {
                if (stream.EOS) //̖֐
                    return new InstructionLine(position, func, null);
                stream.ShiftNext();
                s2 = rowLine.Substring(stream.CurrentPosition, rowLine.Length - stream.CurrentPosition);
                return new InstructionLine(position, func, s2);
            }
            //xł߂łȂ̂ő̂͂BϐŎn܂ĂȂ΂B
            if (!VariableParser.IsVariable(s1))
            {
                errMes = "xEߕÊƂ߂łȂsł";
                goto err;
            }
            if (VariableParser.IsInteger(s1))
                func = BuiltInFunctionCode.SET;
            else
                func = BuiltInFunctionCode.SETS;

            TokenReader.SkipWhiteSpace(stream);//̒P܂œǂݔ΂
            //"ϐ@Zq@"Ƃ`̂͂B
            if ((stream.EOS) || (TokenReader.GetNextTokenType(stream) != TokenType.Operator))
            {
                errMes = "sϐŎn܂Ă܂Zq܂";
                goto err;
            }
            s2 = TokenReader.ReadOperatorString(stream);
            if (s2 == "==")
            {
                if (console != null)
                    console.PrintWarning("Zq\"==\"gĂ܂", position, 0);
                s2 = "=";//"=="Ɏĝ͖{͂\gĂ̂ŎdlɂB
            }
            if (func == BuiltInFunctionCode.SET)
                TokenReader.SkipWhiteSpace(stream);//̑Ȃ玟̒P܂œǂݔ΂
            else if (stream.Current == ' ')
                TokenReader.SkipHalfSpace(stream);//̑Ȃ甼pXy[Xǂݔ΂

            if (stream.EOS)
                s3 = "";
            else
                s3 = rowLine.Substring(stream.CurrentPosition, rowLine.Length - stream.CurrentPosition);


            if (func == BuiltInFunctionCode.SETS)
            {
                if (s2 != "=")
                {
                    errMes = "ϐ̑ɂ͕Zq͎g܂";
                    goto err;
                }
            }
            string right = null;
            if (func == BuiltInFunctionCode.SETS)
                right = s3;
            else
            {
                right = OperatorManager.ReduceAssignmentOperator(s2, s1, s3);
                if (right == null)
                {
                    errMes = "ɎgpłȂZqg܂";
                    goto err;
                }
            }

            return new InstructionLine(position, func, s1, right);
        err:
            return new InvalidLine(position, errMes);
        }

        private static void printWarning(string str, LogicalLine line, int level, bool isError)
        {
            if (isError)
            {
                line.IsError = true;
                line.ErrMes = str;
            }
            if (printWarningLoad == null)
                return;
            printWarningLoad(str, line, level, isError);
        }

        static PrintWarning printWarningLoad = null;
        public static bool SetArgumentTo(InstructionLine line, PrintWarning printW)
        {
            if (line == null)
                return false;
            if (line.Argument != null)
                return true;
            printWarningLoad = printW;
			BuiltInFunctionCode argType = BuiltInFunctionCode.__ARG_MASK__ & line.Function;
            string rowStr = line.ArgumentStr;
            StringStream st = new StringStream(rowStr);
            Argument arg = null;
            IOperandTerm term = null;
            VariableToken varToken = null;
            StringForm strForm = null;
            try
            {
				switch (argType)
                {
					case BuiltInFunctionCode.__ARG_VOID__:
                        if ((rowStr != null) && (rowStr.Length != 0) && (rowStr.Trim().Length != 0))
                            printWarning("svȖ\"" + line.Function.ToString() + "\"Ɉw肳Ă܂", line, 1, false);
                        arg = new VoidArgument();
                        break;
					case BuiltInFunctionCode.__ARG_INT_EXPRESSION__:
                        term = ExpressionParser.ReduceIntegerTerm(st);
                        arg = new ExpressionArgument(term);
                        break;
					case BuiltInFunctionCode.__ARG_STR_EXPRESSION__:
                        term = ExpressionParser.ReduceStringTerm(st);
                        arg = new ExpressionArgument(term);
                        break;
					case BuiltInFunctionCode.__ARG_STR__:
                        if (rowStr == null)
                            rowStr = "";
                        arg = new StrArgument(rowStr);
                        break;
					case BuiltInFunctionCode.__ARG_FORM_STR__:
                        if (rowStr == null)
                            rowStr = "";
                        strForm = new StringForm(rowStr);
						if (Config.Instance.ReduceFormattedStringOnLoad)
                            strForm.Reduce();
                        arg = new FormStrArgument(strForm);
                        break;
					case BuiltInFunctionCode.__ARG_SP_PRINTV__:
                        {
                            List<IOperandTerm> list = new List<IOperandTerm>();
                            while (!st.EOS)
                            {
                                TokenReader.SkipWhiteSpace(st);
								if (st.Current == '\'')
								{
									st.ShiftNext();
									if(!st.EOS && st.Current != ',')
										list.Add(new SingleTerm(TokenReader.ReadStringEndWith(st, null)));
								}
								else
									list.Add(ExpressionParser.ReduceIntegerTerm(st));
                                TokenReader.SkipWhiteSpace(st);
                                if (st.Current == ',')
                                    st.ShiftNext();
                                else
                                    break;
                            }
                            arg = new SpPrintVArgument(list);
                        }
                        break;
					case BuiltInFunctionCode.__ARG_SP_TIMES__:
                        TokenReader.SkipWhiteSpace(st);
                        varToken = VariableParser.ReduceVariable(st);
                        if (varToken == null)
                            printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
                        else if (varToken.IsString)
                            printWarning("\"" + line.Function.ToString() + "\"̑P𕶎ϐɂ邱Ƃ͂ł܂", line, 2, true);
                        else if (varToken.Identifier.Readonly)
                            printWarning("\"" + line.Function.ToString() + "\"̑PɕύXłȂϐw肷邱Ƃ͂ł܂", line, 2, true);

                        TokenReader.SkipWhiteSpace(st);
                        if (st.Current != ',')
                        {
                            if (st.EOS)
                                printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
                            else
                                printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
                            break;
                        }
                        st.ShiftNext();
                        TokenReader.SkipWhiteSpace(st);
                        double d = 0.0;
                        if ((TokenReader.GetNextTokenType(st) != TokenType.Numeric) && (st.Current != '-') && (st.Current != '.'))
                            printWarning("\"" + line.Function.ToString() + "\"̑Qlł͂܂i0Ɖ߂܂j", line, 1, false);
                        else
                        {
                            try
                            {
                                d = ExpressionParser.ReduceDoubleTerm(st);
                            }
                            catch
                            {
                                printWarning("\"" + line.Function.ToString() + "\"̑Qlł͂܂i0Ɖ߂܂j", line, 1, false);
                            }
                        }
                        arg = new SpTimesArgument(varToken, d);
                        break;
					case BuiltInFunctionCode.__ARG_SP_BAR__:
                        {
                            TokenReader.SkipWhiteSpace(st);
                            IOperandTerm value = ExpressionParser.ReduceIntegerTerm(st);
                            TokenReader.SkipWhiteSpace(st);
                            if (st.Current != ',')
                            {
                                printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
                                break;
                            }
                            st.ShiftNext();
                            IOperandTerm max = ExpressionParser.ReduceIntegerTerm(st);
                            TokenReader.SkipWhiteSpace(st);
                            if (st.Current != ',')
                            {
                                printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
                                break;
                            }
                            st.ShiftNext();
                            IOperandTerm length = ExpressionParser.ReduceIntegerTerm(st);
                            TokenReader.SkipWhiteSpace(st);
                            if (st.Current == ',')
                                printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);
                            arg = new SpBarArgument(value, max, length);
                        }
                        break;
					case BuiltInFunctionCode.__ARG_SP_SET__:
                        varToken = VariableParser.ReduceVariable(new StringStream(line.AssignmentDestStr));
                        term = ExpressionParser.ReduceIntegerTerm(st);
                        arg = new SpSetArgument(varToken, term);
                        break;

					case BuiltInFunctionCode.__ARG_SP_SETS__:
                        varToken = VariableParser.ReduceVariable(new StringStream(line.AssignmentDestStr));
                        strForm = new StringForm(rowStr);
						if (Config.Instance.ReduceFormattedStringOnLoad)
                            strForm.Reduce();
                        arg = new SpSetsArgument(varToken, strForm);
                        break;

					case BuiltInFunctionCode.__ARG_SP_SUBSTRING__:
						{
							TokenReader.SkipWhiteSpace(st);
							term = ExpressionParser.ReduceStringTerm(st);
							if (term == null)
								printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								if (st.EOS)
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								else
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							IOperandTerm start = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							IOperandTerm length = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current == ',')
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);

							arg = new SpSubstringArgument(term, start, length);
						}
						break;
					case BuiltInFunctionCode.__ARG_SP_SWAP__:
						{
							TokenReader.SkipWhiteSpace(st);
							IOperandTerm x = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							IOperandTerm y = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current == ',')
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);
							arg = new SpSwapCharaArgument(x, y);
						}
						break;
					case BuiltInFunctionCode.__ARG_SP_VAR__:
						TokenReader.SkipWhiteSpace(st);
						varToken = VariableParser.ReduceVariable(st);
						if (varToken == null)
							printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
						else if ((!varToken.Identifier.IsArray) || (varToken.Identifier.Code == VariableCode.RAND))
							printWarning("\"" + line.Function.ToString() + "\"̑PɔzłȂϐw肷邱Ƃ͂ł܂", line, 2, true);

						TokenReader.SkipWhiteSpace(st);
						if (!st.EOS)
						{
							if (st.Current == ',')
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, true);
							else
								printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 1, true);
							break;
						}
						arg = new SpVarsizeArgument(varToken.Identifier);

						break;
					case BuiltInFunctionCode.__ARG_SP_SAVEDATA__:
						{
							TokenReader.SkipWhiteSpace(st);
							IOperandTerm target = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								if (st.EOS)
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								else
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							term = ExpressionParser.ReduceStringTerm(st);
							if (term == null)
								printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 2, true);
							TokenReader.SkipWhiteSpace(st);
							if (!st.EOS)
							{
								if (st.Current == ',')
									printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, true);
								else
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 1, true);
								break;
							}
							arg = new SpSaveDataArgument(target, term);
							break;
						}
					case BuiltInFunctionCode.__ARG_SP_TINPUTS__:
                        TokenReader.SkipWhiteSpace(st);
                        IOperandTerm time = ExpressionParser.ReduceIntegerTerm(st);
                        TokenReader.SkipWhiteSpace(st);
                        if (st.Current != ',')
                        {
                            if (st.EOS)
                                printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
                            else
                                printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
                            break;
                        }
                        st.ShiftNext();
						TokenReader.SkipWhiteSpace(st);
						term = ExpressionParser.ReduceStringTerm(st);
						if (term == null)
                            printWarning("\"" + line.Function.ToString() + "\"̑QFł܂", line, 2, true);
                        TokenReader.SkipWhiteSpace(st);
                        if (st.Current != ',')
                        {
                            printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
                            break;
                        }
                        st.ShiftNext();
                        IOperandTerm disp = ExpressionParser.ReduceIntegerTerm(st);
                        TokenReader.SkipWhiteSpace(st);
                        if (st.Current == ',')
                            printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);
                        arg = new SpTInputsArgument(time, term, disp);
						break;
					case BuiltInFunctionCode.__ARG_SP_SORTCHARA__:
						{
							varToken = new VariableToken(VariableIdentifier.GetVariableId("NO"), null, null);
							SortOrder order = SortOrder.ASCENDING;
							TokenReader.SkipWhiteSpace(st);
							if (!st.EOS && !st.CurrentEqualTo("FORWARD") && !st.CurrentEqualTo("BACK"))
							{
								varToken = VariableParser.ReduceVariable(st);
								if (varToken == null)
									printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
								else if (!varToken.Identifier.IsCharacterData)
									printWarning("\"" + line.Function.ToString() + "\"̑P̓LN^ϐłȂ΂Ȃ܂", line, 2, true);
								TokenReader.SkipWhiteSpace(st);
								if (st.Current != ',' && !st.EOS)
								{
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
								}
							}
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							if (!st.EOS)
							{
								if (st.CurrentEqualTo("FORWARD"))
									order = SortOrder.ASCENDING;
								else if (st.CurrentEqualTo("BACK"))
									order = SortOrder.DESENDING;
								else
									printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true);
							}

							arg = new SpSortcharaArgument(varToken, order);
							break;
						}

					case BuiltInFunctionCode.__ARG_SP_CALL__:
						{
							TokenReader.SkipWhiteSpace(st);
							List<IOperandTerm> terms = new List<IOperandTerm>();
							string str = TokenReader.ReadStringEndWith(st, null);
							str = str.Trim();
							TokenReader.SkipWhiteSpace(st);
							while (!st.EOS && st.Current == ',')
							{
								st.ShiftNext();
								TokenReader.SkipWhiteSpace(st);
								if (st.EOS || st.Current == ',')
									terms.Add(null);
								else
									terms.Add(ExpressionParser.ReduceExpressionTerm(st));
								TokenReader.SkipWhiteSpace(st);
							}
							IOperandTerm[] args = new IOperandTerm[terms.Count];
							terms.CopyTo(args);
							arg = new SpCallArgment(str, args);
						}
						break;
					case BuiltInFunctionCode.__ARG_SP_CALLFORM__:
						{
							TokenReader.SkipWhiteSpace(st);
							List<IOperandTerm> terms = new List<IOperandTerm>();
							string str = TokenReader.ReadStringEndWith(st, null);
							str = str.Trim();
							strForm = new StringForm(str);
							TokenReader.SkipWhiteSpace(st);
							while (!st.EOS && st.Current == ',')
							{
								st.ShiftNext();
								TokenReader.SkipWhiteSpace(st);
								if (st.EOS || st.Current == ',')
									terms.Add(null);
								else
									terms.Add(ExpressionParser.ReduceExpressionTerm(st));
								TokenReader.SkipWhiteSpace(st);
							}
							IOperandTerm[] args = new IOperandTerm[terms.Count];
							terms.CopyTo(args);
							arg = new SpCallformArgment(strForm, args);
						}
						break;
					case BuiltInFunctionCode.__ARG_SP_FOR_NEXT__:
						{
							TokenReader.SkipWhiteSpace(st);
							varToken = VariableParser.ReduceVariable(st);
							IOperandTerm start = null;
							IOperandTerm end = null;
							IOperandTerm step = null;
							if (varToken == null)
							{ printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true); break; }
							else if (!varToken.Identifier.IsInteger)
							{ printWarning("\"" + line.Function.ToString() + "\"̑P͐l^ϐłȂ΂Ȃ܂", line, 2, true); break; }
							else if (varToken.Identifier.IsCharacterData)
							{ printWarning("\"" + line.Function.ToString() + "\"̑PɃLN^ϐw肷邱Ƃ͂ł܂", line, 2, true); break; }
							else if (varToken.Identifier.Readonly)
							{ printWarning("\"" + line.Function.ToString() + "\"̑P͑\ȕϐłȂ΂Ȃ܂", line, 2, true); break; }
							TokenReader.SkipWhiteSpace(st);
							if (st.EOS)
							{ printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true); break; }
							else if (st.Current != ',')
							{ printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true); break; }
							st.ShiftNext();
							if (st.Current == ',')
								start = new SingleTerm(0);
							else
								start = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.EOS)
							{ printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true); break; }
							else if (st.Current != ',')
							{ printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true); break; }
							st.ShiftNext();
							end = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.EOS)
								step = new SingleTerm(1);
							else if (st.Current == ',')
							{
								st.ShiftNext();
								step = ExpressionParser.ReduceIntegerTerm(st);
							}
							else
							{ printWarning("\"" + line.Function.ToString() + "\"̏ԈĂ܂", line, 2, true); break; }
							arg = new SpForNextArgment(varToken, start, end, step);
							TokenReader.SkipWhiteSpace(st);
							break;
						}

					case BuiltInFunctionCode.__ARG_SP_POWER__:
						{
							TokenReader.SkipWhiteSpace(st);
							varToken = VariableParser.ReduceVariable(st);
							if (varToken == null)
								printWarning("\"" + line.Function.ToString() + "\"̑PFł܂", line, 2, true);
							else if (varToken.IsString)
								printWarning("\"" + line.Function.ToString() + "\"̑P𕶎ϐɂ邱Ƃ͂ł܂", line, 2, true);
							else if (varToken.Identifier.Readonly)
								printWarning("\"" + line.Function.ToString() + "\"̑PɕύXłȂϐw肷邱Ƃ͂ł܂", line, 2, true);

							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							IOperandTerm x = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current != ',')
							{
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 2, true);
								break;
							}
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							IOperandTerm y = ExpressionParser.ReduceIntegerTerm(st);
							TokenReader.SkipWhiteSpace(st);
							if (st.Current == ',')
								printWarning("\"" + line.Function.ToString() + "\"̈܂", line, 1, false);

							arg = new SpPowerArgument(varToken, x, y);
							break;
						}
                    default:
                        printWarning("ݒ肳ĂȂ^", line, 2, true);
                        return false;
                }
            }
            catch (CodeEE e)
            {
                printWarning(e.Message, line, 2, true);
                return false;
            }
            if (arg == null)
            {
                if (!line.IsError)
                    printWarning("߂̈͒ɓłȂG[", line, 2, true);
                return false;
            }
            line.Argument = arg;
            return true;
        }
    }
}
